from . import device
import os, threading, socket, subprocess, time


_minitouches = {}


class CONST:
    PACKAGEDIR = os.path.dirname(os.path.abspath(__file__))
    PREBUILDDIR = os.path.join(PACKAGEDIR,'prebuild/minitouch/')
    DEFAULTPARAMS = { 'port': 15544, 'frq':120}
    COMMIT = 'c'
    RESETALL = 'r'
    TOUCHUP = 'u'
    TOUCHDOWN = 'd'
    TOUCHMOVE = 'm'
    WAIT = 'w'
                    

def getMinitouch(device_id = None, **params):
    global _minitouches
    
    if device_id in _minitouches:
        _minitouches[device_id].updateParams(**params)
        return _minitouches[device_id]
        
    device_id = device.getDevice(device_id).id    
    if device_id in _minitouches:
        _minitouches[device_id].updateParams(**params)
        return _minitouches[device_id]
        
    newtouch = _Minitouch(device_id, **params)
    _minitouches[device_id] = newtouch
        
    return _minitouches[device_id]


class _Client(threading.Thread):

    def __init__(self, device_id, port, max_cache_size=255):
        super().__init__()
        self.id = device_id
        self.port = port
        self.header = None
        self.cache = []
        self.stopped = True
        self.iserror = False
        self.conn = None
        self.condition = threading.Condition()
    

    def __parseHeader(self):
    
        headertext = self.conn.recv(1024).decode('utf-8')
        while len(headertext.split('\n'))<3:
            headertext += self.conn.recv(1024).decode('utf-8')
        _header = {}
        lines = headertext.split('\n')
        #print(lines)
        for line in lines:
            l = line.split()
            if len(l)<=0:
                continue
            if l[0] == 'v':
                _header['version'] = int(l[1])
            elif l[0] == '^':
                _header['max-contacts'] = int(l[1])
                _header['max-x'] = int(l[2])
                _header['max-y'] = int(l[3])
                _header['max-pressure'] = int(l[4])
            elif l[0] == '$':
                _header['pid'] = int(l[1])
        self.header = _header
        return _header

    
    def _sendRaw(self, raw):
        #print(raw)
        return self.conn.send(raw.encode('utf-8'))
        
    
    def xyDecode(self, x, y):
        # 比例坐标转为触控坐标
        if x>=0 and x<=1 and y>=0 and y<=1:
            x, y = x*self.header['max-x'], y*self.header['max-y']
        return int(x), int(y)
    
    def pressureLimit(self, pressure=None):
        pressure = self.header['max-pressure'] if pressure is None or pressure>self.header['max-pressure'] else pressure
        return int(pressure)
    
    def _send(self, head, contact_or_ms=None, x=None, y=None, pressure=None):
        
        #print(f'client_send:{head} {contact_or_ms} {x} {y} {pressure}')
        
        if head in [CONST.COMMIT, CONST.RESETALL]:
            cmd = f'{head}\n'
        elif head in [CONST.WAIT, CONST.TOUCHUP]:
            contact_or_ms = 0 if contact_or_ms is None else contact_or_ms
            cmd = f'{head} {int(contact_or_ms)}\n'
        elif head in [CONST.TOUCHDOWN, CONST.TOUCHMOVE] and x is not None and y is not None:
            pressure = self.pressureLimit(pressure)
            x, y = self.xyDecode(x, y)
            cmd = f'{head} {int(contact_or_ms)} {x} {y} {pressure}\n'
        else:
            return False
        return self._sendRaw(cmd)        
    
    
    def notify(self):
        with self.condition:
            self.condition.notify_all()
    
    
    def send(self, *params):
        # input: head, contact_or_ms, px, py, pressure
        # sample:
        # 'c'
        # 'd', 0, 0.1, 0.1, 50
        self.cache.append(params)
        self.notify()
    
    
    def __sendCache(self):
        
        sendlen = len(self.cache)
        
        for i in range(sendlen):
            self._send(*self.cache[i])
            
        self.cache = self.cache[sendlen:]
        return sendlen
        
    
    def run(self):
        
        # todo 重复打开端口错误
        self.conn = socket.socket()
        self.conn.connect(('127.0.0.1',self.port)) 
        # todo 判断端口是否正确连接并采取相应措施
        
        self.__parseHeader()
        
        self._send(CONST.RESETALL)
        
        self.stopped = False
        
        while not(self.stopped):
            with self.condition:
                self.condition.wait()
            while len(self.cache)>0:
                self.__sendCache()
            
        self.conn.close()
        
    
class _Contact:

    def __init__(self, father, contact_id):
        self.id = contact_id
        self.father = father
        self.islocked = False
        self.isremoved = False
        self.sendcache = []
        self.condition = threading.Condition()
        self.idlecondition = threading.Condition()
        self.framedurationts = 1./father.params['frq']
        self.x = None
        self.y = None
        self.pressed = False
        self.supervisor = threading.Thread(target=self.supervisorThread)
        self.supervisor.start()
        
        
    def __enter__(self):
        self.islocked = True
        return self
        
        
    def __exit__(self, exc_type, exc_value, traceback):
        self.islocked = False
        
    
    @property
    def inuse(self):
        return self.islocked or self.isbusy
    
    
    @property
    def isbusy(self):
        return len(self.sendcache)>0
        
    
    def supervisorThread(self):
        status = ''
        wait_until_ts = None
        while not self.isremoved:
            
            if status == '' and self.isbusy:
                params = self.sendcache[0]
                self.sendcache = self.sendcache[1:]
                
                if params[0]==CONST.WAIT:
                    status = 'wait'
                    wait_until_ts = time.time()+int(params[1])/1000
                
                elif params[0] == CONST.COMMIT:
                    pass
                    
                else:
                    self._send(*params)
                    self._send(CONST.COMMIT)
                    time.sleep(self.framedurationts)
                    
            elif status == 'wait':
                if time.time()>=wait_until_ts:
                    status = ''
                else:
                    time.sleep(self.framedurationts)

            else:
                with self.idlecondition:
                    self.idlecondition.notify_all()
                    
                with self.condition:
                    self.condition.wait()
                
                
    def notify(self):
        with self.condition:
            self.condition.notify_all()
    
    
    def clearCache(self):
        self.sendcache = []
            
        
    def close(self):
        
        self.clearCache()
        with self.idlecondition:
            self.idlecondition.notify_all()
            
        self._send(CONST.TOUCHUP, self.id)
        
        self.islocked = False
        
    
    def remove(self):
        self.isremoved = True
        self.notify()
        self.close()
        self.supervisor.join()
        
    
    def xyEncode(self, x, y):
        return self.father.xyEncode(x, y)
        '''
        if x>1 or y>1:
            screensize = self.father.screensize
            x , y = x/screensize['width'], y/screensize['height']
        return x, y
        '''
        
    def xyLimit(self, x, y):
        x = 1 if x>1 else x
        y = 1 if y>1 else y
        x = 0 if x<0 else x
        y = 0 if y<0 else y 
        return x, y
    
    
    def _send(self, *params):
        # input: head, contact_or_ms, px, py, pressure
        # sample:
        # 'c'
        # 'd', 0, 0.1, 0.1, 50
        if len(params)<=0:
            return
        if len(params)>=4:
            params = list(params)
            params[2:4] = self.xyLimit(*self.xyEncode(*params[2:4]))
        head = params[0]
        if head in [CONST.TOUCHDOWN]:
            self.pressed = True
            self.x , self.y = params[2:4]
        elif head in [CONST.TOUCHMOVE]:
            self.x , self.y = params[2:4]
        elif head in [CONST.RESETALL, CONST.TOUCHUP]:
            self.pressed = False   
        
        #print(params)
        return self.father.send(*params)
    

    def _sendBackground(self, *params):
        self.sendcache.append(params)
        self.notify()
        return len(self.sendcache)
        
        
    def _sendChain(self, chain, background=True):
        for params in chain:
            self._sendBackground(*params)
        self.notify()
        if not(background):
            with self.idlecondition:
                self.idlecondition.wait()
        return True
        #results = [self._send(*params) for params in chain]
        #return all(results)
    
    
    def _updateChain(self, chain, background=True):
        self.clearCache()
        return self._sendChain(chain, background)
    
    
    def down(self, x=None, y=None, pressure=None, background = True):
        if (x is None or y is None):
            if self.x is not None and self.y is not None:
                x, y = self.x, self.y
            else:
                return False
                
        chain = [[CONST.TOUCHDOWN , self.id, x, y, pressure], [CONST.COMMIT]]
        
        return self._updateChain(chain, background)
        
    
    def up(self, background = True):
        
        chain = [[CONST.TOUCHUP , self.id], [CONST.COMMIT]]
        
        return self._updateChain(chain, background)
        
    
    def moveto(self, x, y, pressure=None, background = True ):
    
        chain = [[CONST.TOUCHMOVE ,self.id, x, y, pressure], [CONST.COMMIT]]
        
        return self._updateChain(chain, background)
                
        
    def move(self, offsetx, offsety, pressure=None, background = True):
        if self.x is None or self.y is None:
            return False
            
        offsetx, offsety = self.xyEncode(offsetx, offsety)    
        x, y = self.x+offsetx, self.y+offsety
        
        chain = [[CONST.TOUCHMOVE ,self.id, x, y, pressure], [CONST.COMMIT]]
        
        return self._updateChain(chain, background)
        
    '''    
    def _wait(self, ms=10, background = True):
        
        params = [CONST.WAIT, ms]
        
        if background:
            self._sendBackground(*params)
            self.notify()
            return True
            
        return self._send(*params)
    '''
    
    def tap(self, x, y, pressure=None, duration_ms=50, background = True):
        # 简化点击指令
        chain = [
            [CONST.TOUCHDOWN , self.id, x, y, pressure],
            [CONST.COMMIT],
            [CONST.WAIT, duration_ms],
            [CONST.TOUCHUP , self.id],
            [CONST.COMMIT]
            ]
        
        return self._updateChain(chain, background)        
        
    
    def swipe(self, x1, y1, x2, y2, pressure=None, duration_ms=200, background=True):
        frames = int(duration_ms/1000/self.framedurationts)
        xp = (x2-x1)/frames
        yp = (y2-y1)/frames
        
        chain = [[CONST.TOUCHDOWN , self.id, x1, y1, pressure], [CONST.COMMIT]]
        for i in range(1,frames):
            chain.append([CONST.WAIT, self.framedurationts])
            chain.append([CONST.TOUCHMOVE ,self.id, x1+xp*i, y1+yp*i, pressure])
            chain.append([CONST.COMMIT])
        
        chain.append([CONST.TOUCHUP , self.id])
        chain.append([CONST.COMMIT])
        
        return self._updateChain(chain, background)

    
    
class _Minitouch:
    
    def __init__(self, device_id=None, autorotation=True, **params):
        self.device = device.getDevice(device_id)
        self.id = self.device.id
        self.params = dict(CONST.DEFAULTPARAMS, **params)
        self.isinstalled = False
        self.isensured = False
        self.server = None
        self.client = None
        self.orientation = None
        self.__screensize = None
        self.autorotation = autorotation
        self.contacts = []
        self.stopping = False
        self.default_contact = None
        if autorotation:
            self.device.supervisorAdd('Orientation')
            self.device.supervisorAdd('WMSize')
            
            
    def install(self):
        if self.isinstalled:
            return self.params['remote_bin_path'], self.params['remote_dir']
        print('正在部署minitouch。。。', end='')
        dev = self.device
        self.__stopServer()
        abi = dev.abi
        sdk = dev.sdk
        binName = 'minitouch' if int(sdk)>=16 else 'minitouch-nopie'
        binPath = os.path.join(CONST.PREBUILDDIR, f'{abi}/bin/{binName}')
        remoteDir = dev.tmpdir
        remoteBinPath = dev.pushTmp(binPath)
        dev.shell(f'chmod 777 {remoteBinPath}')
        self.params['remote_bin_path'] = remoteBinPath
        self.params['remote_dir'] = remoteDir
        print('ok')
        self.isinstalled = True
        return remoteBinPath, remoteDir
        
    
    @property
    def isServerRun(self):
        return self.server and self.server.poll() is None
    
    @property
    def screensize(self):
        return self.device.supervisor.params['WMSize'] if self.autorotation else self.__screensize
    
    def start(self):
        self.__startServer()
        self.__startClient()
        self.default_contact = self.useContact()
        
        
        
    def stop(self):
        if self.stopping:
            if self.server is not None:
                self.server.wait()
            self.stopping = False
            return
        self.stopping = True    
        self.__stopClient()
        self.__stopServer()
        self.stopping = False
        
        
    def close(self):
        self.stop()
        
        
    def restart(self):
        self.stop()
        self.start()
        

    def updateParams(self, **params):
        thesame = [v==self.params.get(k,None) for k,v in params.items()]
        if len(thesame)>0 and not all(thesame):
            self.params = dict(self.params, **params)
            self.restart()
    

    def __startClient(self):
        if self.client is not None and self.client.is_alive():
            return 
        self.client = _Client(self.id, self.params['port'])
        self.client.start()
        while self.client.stopped:
            time.sleep(1)
        self.contacts = [_Contact(self, i) for i in range(self.client.header['max-contacts'])]
        
        
    def __stopClient(self):
        if self.client is None or not(self.client.is_alive()):
            return
        for contact in self.contacts:
            contact.remove()
        self.client.stopped = True
        self.client.notify()
        self.client.join()
    

    def __stopServer(self):
        self.device.shell('pkill minitouch')
        if self.server is None:
            return
        self.server.terminate()
        self.server.wait()
    
    
    def __startServer(self):
        if self.isServerRun:
            return 
        remoteBinPath, remoteDir = self.install()
        self.orientation = self.device.getOrientation()
        self.__screensize = self.device.getWMSize()
        port = self.params['port']
        self.server = self.device.shell(remoteBinPath, readout=False)
        while True:
            if 'contacts' in self.server.stderr.readline().decode().lower():
                break        
            # todo: event无写入权限的处理（需要root）    
        self.device.adb(f'forward tcp:{port} localabstract:minitouch')
        
    
    def xyEncode(self, x, y):
        # 统一为比例坐标
        # 输入坐标使用比例坐标，或按照屏幕分辨率而非触控分辨率
        orientation = self.device.supervisor.params['Orientation'] if self.autorotation else self.orientation
        if x>1 or y>1:
            if orientation in [1,3]:
                x , y = x/self.screensize['height'], y/self.screensize['width']
            else:
                x , y = x/self.screensize['width'], y/self.screensize['height']
        
        return x, y
    

    def xyTransform(self, x, y):
        # 输入比例坐标
        # 使屏幕左上角始终为坐标原点
        # orientation*90为屏幕顺时针旋转度数
        orientation = self.device.supervisor.params['Orientation'] if self.autorotation else self.orientation
        
        if orientation == 1:
            x, y = 1-y, x
            
        elif orientation == 2:
            x, y = 1-x, 1-y 
            
        elif orientation ==3:
            x, y = y, 1-x
        
        return x, y
    
        
    def ensure(self):
        if self.client is None or not(self.client.is_alive()):
            self.restart()
        
    def send(self, *params):
        # input: head, contact_or_ms, px, py, pressure
        # sample:
        # 'c'
        # 'd', 0, 0.1, 0.1, 50
        self.ensure()
        if len(params)>=4:
            params = list(params)
            params[2:4] = self.xyTransform(*self.xyEncode(*params[2:4]))
        return self.client.send(*params)
    
    
    def useContact(self, cid=None):
        self.ensure()
        if cid is not None:
            return self.contacts[cid]
        for contact in self.contacts[::-1]:
            if not(contact.inuse):
                contact.islocked = True
                return contact
        print('所有contact均被占用，请使用with ... as 或 使用后close()以释放')
        return None
    
    
    def down(self, x=None, y=None, pressure=None, background = True):
        self.ensure()
        return self.default_contact.down(x=x, y=y, pressure=pressure, background=background)

        
    
    def up(self, background = True):
        
        return self.default_contact.up(background=background)
        
    
    def moveto(self, x, y, pressure=None, background = True ):
    
        return self.default_contact.moveto(x=x, y=y, pressure=pressure, background=background)
                
        
    def move(self, offsetx, offsety, pressure=None, background = True):
    
        return self.default_contact.move(x=offsetx, y=offsety, pressure=pressure, background=background)


from flask import Flask, render_template, Response, request
from werkzeug.serving import make_server
import threading, os
from .. import minicap, minitouch, ime, version, device
from socketman import WebsocketServer
import json

# todo: 无设备连接时需要错误提示

class CONST:
    PACKAGEDIR = os.path.dirname(os.path.abspath(__file__))
    HTMLDIR = os.path.join(PACKAGEDIR,'templates/')
    HOMEPAGENAME = 'index.html'
    DEFAULTPARAMS = {'host':'127.0.0.1', 'port':5000, 'device_id':None, 'frame_provider': None }



class WebUI(threading.Thread):
    
    def __init__(self, **params):
        super().__init__()
        self.params = dict(CONST.DEFAULTPARAMS, **params)
        self.device = device.getDevice(self.params['device_id'])
        self.params["device_id"] = self.device.id
        self.minicap = minicap.getMinicap(self.params['device_id'])
        self.minitouch = minitouch.getMinitouch(self.params['device_id'])
        self.ime = ime.getIME(self.params['device_id'])
        self.contact = None
        self.id = self.minicap.id if self.minicap is not None else '未连接'
        self.frameprovider = lambda :self.minicap.cap( captype=minicap.CONST.JPG, sync=True) if self.params['frame_provider'] is None else self.params['frame_provider']
        self.stopped = False
        app = Flask(__name__)
        self.app = app
        self.server = make_server(
            host = self.params['host'], 
            port = self.params['port'], 
            app = app,
            threaded = True
            )
        self.url = f'http://localhost:{self.params["port"]}'
        
        self.websocket_server = WebsocketServer(host='localhost', port = self.params['port']+1)
        
        @self.websocket_server.ON_RECV
        def recv_handler(txt, wsid):
            print(txt)
            try:
                switchchanges = json.loads(txt)
                for k, v in switchchanges.items():
                    #print(f'{k}: {v}')
                    if k == 'running' and v==False:
                        t = threading.Thread(target = self.stop)
                        t.start()
                    elif k== 'action':
                        self.doAction(*v)
                    elif k== 'button':
                        self.doButton(v)
                    elif k== 'string':
                        self.doInput(v)
                        #self.stop()
                    #print(f'{k}: {v}')
                    #self.switches[k] = v
                    #if k in ['pause', 'running']:
                    #    with self.switches['pause_condition']:
                    #        self.switches['pause_condition'].notify_all()
            except Exception as e:
                print(e)
                pass
            
        
                
        @app.route('/')
        def homepage():
            #if os.path.exists(CONST.HOMEPAGE):
            webvars = {'device_id':self.id, 'websocket_url':self.websocket_server.url, 'version' : version.version}
            print(webvars)
            return render_template(CONST.HOMEPAGENAME, **webvars)
            return 'homepage lost.'
        

        @app.route('/check')
        def check():
            return 'hello world.'
            
        
        @app.route('/video_feed')
        def video_feed():
            def wrapcap():
                while not(self.stopped):
                    frame = self.frameprovider()
                    yield (
                        b'--frame\r\n'
                        b'Content-Type: image/jpeg\r\n\r\n' + frame + b'\r\n'
                        )
            return Response(
                wrapcap(), 
                mimetype='multipart/x-mixed-replace; boundary=frame'
                )

    def doInput(self, str):
        self.ime.input(str)
        pass

    def doButton(self, key):
        # todo： 发送按钮，back， home， menu
        key = key.lower()
        if key=='back':
            self.device.shell('input keyevent KEYCODE_BACK')
        elif key=='home':
            self.device.shell('input keyevent KEYCODE_HOME')
        elif key=='menu':
            self.device.shell('input keyevent KEYCODE_MENU')
        elif key=='power':
            self.device.shell('input keyevent KEYCODE_POWER')
        elif key=='volume_up':
            self.device.shell('input keyevent KEYCODE_VOLUME_UP')
        elif key=='volume_down':
            self.device.shell('input keyevent KEYCODE_VOLUME_DOWN')
        elif key=='volume_mute':
            self.device.shell('input keyevent KEYCODE_VOLUME_MUTE')
        elif key=='camera':
            self.device.shell('input keyevent KEYCODE_CAMERA')
        elif key=='camera_focus':
            self.device.shell('input keyevent KEYCODE_FOCUS')
        elif key=='scale':
            self.device.shell('input keyevent KEYCODE_SCALE')
        pass

    def doAction(self, a, x, y):
        if a=='down':
            self.contact.down(x, y)
        elif a=='up':
            self.contact.up()
        elif a=='move':
            self.contact.moveto(x, y)
    
    
    def run(self):
        self.websocket_server.start()
        #print('socket start')
        self.contact = self.minitouch.useContact()
        self.server.serve_forever()
        '''    
        self.app.run(
            host=self.params['host'], 
            port=self.params['port'], 
            debug=False, 
            use_reloader=False
            )
        '''   
        
    def stop(self):
        #import requests
        #requests.get(f'http://localhost:{self.params["port"]}/shutdown')
        self.stopped = True
        self.server.shutdown()
        #print('wsstopped')
        self.join()
        #print('joined')
        self.websocket_server.stop()    
        #print('socketstopped')
        self.minicap.close()
        #self.contact.close()
        self.minitouch.close()
        #print('minicap_stopped')


from . import device
import os, base64

# todo: 输入法close后还原

_imes = {}

class CONST:
    PACKAGEDIR = os.path.dirname(os.path.abspath(__file__))
    PREBUILDAPK = os.path.join(PACKAGEDIR,'prebuild/ADBKeyboard.apk')
    PACKAGENAME = 'com.android.adbkeyboard'
    IMENAME = 'com.android.adbkeyboard/.AdbIME'

def getIME(device_id = None):
    global _imes
    if device_id in _imes:
        return _imes[device_id]
        
    device_id = device.getDevice(device_id).id
    if device_id in _imes:
        return _imes[device_id]
        
    newime = _IME(device_id)
    _imes[device_id] = newime
    return _imes[device_id]
    

class _IME:

    def __init__(self, device_id=None):
        self.device = device.getDevice(device_id)
        self.id = self.device.id
        self.isinstalled = False
        self.isensured = False

        
    def install(self):
        if self.isinstalled:
            # Raise
            print('ADBkeyboard安装失败，可能无法正常工作。')
            return 
        print('正在安装ADBkeyboard。。。', end='')
        if not self.device.installAPK(CONST.PREBUILDAPK, packagename = CONST.PACKAGENAME):
            # Raise
            pass
        self.device.shell(f'ime enable {CONST.IMENAME}')
        print('ok')
        self.isinstalled = True
        
    
    def getEnabled(self):
        imeEnabled = self.device.shell('ime list -s')
        return CONST.IMENAME in imeEnabled
        
    
    def getSelected(self):
        currentIme = self.device.shell('settings get secure default_input_method').strip()
        return CONST.IMENAME in currentIme
    
    
    def getActive(self):
        return self.device.getIMEInputActive()
    
    
    def select(self):
        dev = self.device
        dev.shell(f'ime set {CONST.IMENAME}')
        
    
    def ensure(self):
        if not(self.getSelected()):
            if not(self.getEnabled()):
                self.install()
            self.select()
        self.isensured = True
    
    
    def b64Encode(self, txt):
        return str(base64.b64encode(txt.encode('utf-8')))[1:]
    
    
    def input(self, txt):
        if not self.isensured:
            self.ensure()
        return  self.device.shell(f'am broadcast -a ADB_INPUT_B64 --es msg {self.b64Encode(txt)}', readout=False)
        
    
    def inputSafe(self, txt):
        self.ensure()
        return  self.device.shell(f'am broadcast -a ADB_INPUT_B64 --es msg {self.b64Encode(txt)}', readout=True)


import subprocess
import sys, os
from .common import ParamSupervisor
from .adb import ADB_PATH

_devices = {}

# todo 检查设备在线状态

class CONST:
    STATUS_AVAILABLE = True
    TABLET_720P = {'width':1280, 'height':720, 'density':240}
    


def listDevicesL():
    cmd = f'{ADB_PATH} devices -l'
    subp = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE)
    rl = subp.stdout.readlines()
    def parseDevicesList(rawlines):
        devicelist = {}
        for line in rawlines:
            lineitems = line.decode().split()
            if len(lineitems)>2 and lineitems[1].lower() in ['device']:
                devicelist[lineitems[0]] = lineitems[1:]
        return devicelist
    return parseDevicesList(rl)


def listDevices():
    return listDevicesL()
    

def connectDevice(ip='127.0.0.1',port='16416'):
    cmd = f'{ADB_PATH} connect {ip}:{port}'
    subp = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE)
    results = subp.stdout.read().decode().split()
    return results[0].lower() in ['connected','already']
    
    
def getDevice(device_id=None):
    global _devices
    if device_id not in _devices:
        devicelist = listDevicesL()
        if device_id is None and len(devicelist)>=1:
            device_id = list(devicelist.keys())[0]
        elif device_id not in devicelist:
            if len(devicelist)==0:
                print(f'未检测到任何通过adb连接的设备，请先连接')
            else:
                print(f'指定安卓设备{device_id}未通过adb连接，请先连接')
            return None
        _devices[device_id] = _Device(device_id = device_id, header = devicelist[device_id])
    return _devices[device_id]




class _Device:
    
    
    def __init__(self, device_id = None, header = None):
        self.id = device_id
        self.isenable = True
        self.tmpdir = '/data/local/tmp/'
        self.header = header
        self.wmphysical = None
        self.supervisor = ParamSupervisor(self, stop_callback=self.remove)
        self.isenable = True
        
    
    def remove(self):
        global _devices
        print('设备停止工作')
        self.isenable = False
        _devices.pop(self.id)
        
        
    def supervisorStart(self):
        if self.supervisor is not None and self.supervisor.is_alive():
            return 
        self.supervisor.start()
        
        
    def supervisorStop(self):
        if self.supervisor is None or not(self.supervisor.is_alive()):
            return
        self.supervisor.stopped = True
        self.supervisor.join()
        
        
    def supervisorAdd(self, param):
        self.supervisor.addParam(param)
        self.supervisorStart()
        
        
    @staticmethod
    def pathjoin(*paths):
        pathitems = []
        for path in paths:
            pathitems+=path.split('/')
        return '/'.join(pathitems)
        
        
    @property    
    def adbhead(self):
        if self.id:
            return f'{ADB_PATH} -s {self.id} '
        return f'{ADB_PATH} '
        
    
    def getStatus(self):
        return self.getOnline()
    
    
    def requireRoot(self):
        self.adb('root', readout=True)
        
    
    def adb(self, cmd, readout=True):
        acmd = self.adbhead+cmd
        subp =  subprocess.Popen(acmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, stdin=subprocess.PIPE)
        if readout:
            return subp.stdout.read().decode('utf-8').strip()
        return subp
        
        
    def shell(self, cmd, readout=True):
        return self.adb(f'shell "{cmd}"', readout)
        
    
    def getOnline(self):
        return self.id in listDevicesL()
    
    
    def getOrientation(self):
        #0 原始竖屏， 1 左横屏， 3 右横屏
        key = 'orientation='
        s = self.shell(f"dumpsys display | grep \'{key}\' | head -1", readout=True)
        i = s.find(key)
        oid = i+len(key)
        if (len(s)<oid):
            return 0
        return int(s[oid])
        
        
    def getIMEInputActive(self):
        # 检测输入框是否存在 或 当前是否可以使用输入法输入
        return self.shell("dumpsys input_method | grep mServedInputConnectionWrapper", readout=True).strip().split('=')[-1].lower() !='null'

    
    @property    
    def ABI(self):
        pname =f'_{sys._getframe().f_code.co_name}'
        if pname not in vars(self):
            vars(self)[pname] = self.shell("getprop ro.product.cpu.abi", readout=True).strip()
        return vars(self)[pname]
    
    
    @property
    def abi(self):
        return self.ABI
            
            
    @property    
    def SDK(self):
        pname =f'_{sys._getframe().f_code.co_name}'
        if pname not in vars(self):
            self.__dict__[pname] = int(self.shell("getprop ro.build.version.sdk", readout=True).strip())
        return vars(self)[pname]      


    @property
    def sdk(self):
        return self.SDK

        
    def installAPK(self, apkpath, packagename=None):
    
        if os.path.splitext(apkpath)[-1].lower() != '.apk':
            return False
            
        tmppath = self.pushTmp(apkpath)
        
        if packagename:
            cmd = f'pm install -r -i {packagename} {tmppath}'
        else:
            cmd = f'pm install -r {tmppath}'
            
        self.shell(cmd)
        
        return True
    
    
    def pushTmp(self, fpath):
        fname = os.path.basename(fpath)
        tmppath = self.pathjoin(self.tmpdir, fname)
        self.adb(f'push {fpath} {tmppath}')
        return tmppath
    
    
    def getWMSize(self):
        wmSizes = [l.lower() for l in self.shell('wm size').split('\r\n')]
        overideSize = None
        for line in wmSizes:
            if 'physical' in line:
                physicalSize = [int(x) for x in line.split()[-1].split('x')]
            elif 'override' in line:
                overideSize = [int(x) for x in line.split()[-1].split('x')]
        if self.wmphysical is None:
            self.wmphysical = {}
        self.wmphysical['width'], self.wmphysical['height'] = physicalSize
        if overideSize:
            wh =  overideSize
        else:
            wh = physicalSize
        return {'width':wh[0], 'height':wh[1]}
    
    
    def getWMDensity(self):
        wmDensitys = [l.lower() for l in self.shell('wm density').split('\r\n')]
        overideDensity = None
        for line in wmDensitys:
            if 'physical' in line:
                physicalDensity = int(line.split()[-1])
            elif 'override' in line:
                overideDensity = int(line.split()[-1])
        if self.wmphysical is None:
            self.wmphysical = {}
        self.wmphysical['density'] = physicalDensity
        if overideDensity:
            density = overideDensity
        else:
            density = physicalDensity
        return {'density': density}
    
    
    def getWM(self):
        wmSize = self.getWMSize()
        wmDensity = self.getWMDensity()
        return dict(wmSize, **wmDensity)
        
        
    def setWM(self, width, height, density = None, fit_orientation = True):
        if fit_orientation: # 确保方向与设备的物理屏幕一致
            pwm = self.wmphysical if self.wmphysical else self.getWMSize()
            width, height = (width, height) if ((width/height)>1) ^ ((pwm['width']/pwm['height'])<=1) else (height, width)
        self.shell(f'wm size {width}x{height}')
        if density:
            self.shell(f'wm density {density}')
            
            
    def resetWM(self):
        self.shell(f'wm size reset')
        self.shell(f'wm density reset')
        
        
    


from PIL import Image
import cv2

def imshow(imatbgr):
    if len(imatbgr.shape)==3:
        imat = cv2.cvtColor(imatbgr, cv2.COLOR_BGR2RGB)
    else:
        imat = imatbgr
    img = Image.fromarray(imat)
    display(img)


def drawrect(imat, rect, color=(0,0,255)):
    return cv2.rectangle(imat, rect[:2], rect[2:], color)
    
    
def drawyrect(imat, lmrect, color=(0,0,255)):
    h,w,c = imat.shape
    cp = [w*lmrect[0], h*lmrect[1]]
    wh = [w*lmrect[2], h*lmrect[3]]
    rect = [(int(cp[0]-wh[0]/2), int(cp[1]-wh[1]/2)), (int(cp[0]+wh[0]/2), int(cp[1]+wh[1]/2))]   
    result = cv2.rectangle(imat, *rect, color)
    return result
    
    



import os
from . import device, common
import threading, time, cv2
import socket
import numpy as np

_minicaps={}

# todo: 
#   传输数据损坏或长时间无数据更新时需要重启
#   设备离线时需要停止服务


class CONST:
    JPG = 0
    IMG = 1
    CV2 = 2
    PACKAGEDIR = os.path.dirname(os.path.abspath(__file__))
    PREBUILDDIR = os.path.join(PACKAGEDIR,'prebuild/minicap/')
    DEFAULTPARAMS = { 'port': 15534, 'maxfps':15, 'jpeg_quality':80, 
                    'cache_size':60, 'cache_types':[JPG]}
    
    
    
def getMinicap(device_id = None, **params):
    global _minicaps
    
    if device_id in _minicaps:
        _minicaps[device_id].updateParams(**params)
        return _minicaps[device_id]
        
    device_id = device.getDevice(device_id).id    
    if device_id in _minicaps:
        _minicaps[device_id].updateParams(**params)
        return _minicaps[device_id]
        
    newcap = _Minicap(device_id, **params)
    _minicaps[device_id] = newcap
        
    return _minicaps[device_id]
    

    
class _Client(threading.Thread):

    def __init__(self, device_id, port, cache_size=30, cache_types = [CONST.JPG, CONST.CV2]):
        super().__init__()
        self.id = device_id
        self.port = port
        self.buff = b''
        self.header = None
        self.cache = common.Cache(maxsize = cache_size)
        self.cachetypes = cache_types
        self.stopped = True
        self.iserror = False
        self.conn = None
        self.capcount = 0
        self.framecondition = threading.Condition()
        
    def __parseHeader(self):
        HEADLEN = 24
        self.__collectBuff(HEADLEN)
        buff = self.buff
        
        seekl, seekr = 0, HEADLEN
        hbuff = self.buff[seekl:seekr]
        ''' 0 	1 	unsigned char 	Version (currently 1)
        1 	1 	unsigned char 	Size of the headerer (from byte 0)
        2-5 	4 	uint32 (low endian) 	Pid of the process
        6-9 	4 	uint32 (low endian) 	Real display width in pixels
        10-13 	4 	uint32 (low endian) 	Real display height in pixels
        14-17 	4 	uint32 (low endian) 	Virtual display width in pixels
        18-21 	4 	uint32 (low endian) 	Virtual display height in pixels
        22 	1 	unsigned char 	Display orientation
        23 	1 	unsigned char 	Quirk bitflags (see below)'''
        width, height = np.frombuffer(hbuff[14:18],dtype=np.uint32).squeeze(), np.frombuffer(hbuff[18:22],dtype=np.uint32).squeeze()
        rotation = np.frombuffer(hbuff[0:1],dtype=np.uint8).squeeze()
        header = {
            'version': np.frombuffer(hbuff[0:1],dtype=np.uint8).squeeze(),
            'headerer_size': np.frombuffer(hbuff[1:2],dtype=np.uint8).squeeze(),
            'minicap_pid': np.frombuffer(hbuff[2:6],dtype=np.uint32).squeeze(),
            'real_width': np.frombuffer(hbuff[6:10],dtype=np.uint32).squeeze(),
            'real_height': np.frombuffer(hbuff[10:14],dtype=np.uint32).squeeze(),
            'virtual_width': width,
            'virtual_height': height,
            'rotation': rotation,
            'quirk': np.frombuffer(hbuff[0:1],dtype=np.uint8).squeeze(),
            'width': width if rotation==0 else height,
            'height': height if rotation==0 else width
        }
        self.header = header
        self.buff = self.buff[HEADLEN:]
        return header
    
    
    def __collectBuff(self, req_len=0):
        try:
            while len(self.buff)<req_len:
                new_buff = self.conn.recv(req_len)
                self.buff += new_buff
        except:
            pass
        return self.buff
    
    
    def __parseFrame(self):
        self.__collectBuff(4)
        req_len = np.frombuffer(self.buff[:4],dtype=np.uint32).squeeze()
        self.buff = self.buff[4:]
        self.__collectBuff(req_len)
        with self.cache.withnew as newcache:
            
            jpgbuff = self.buff[:req_len]
            
            newcache['ts'] = time.time()
            newcache[CONST.JPG] = jpgbuff
            
            for ctype in self.cachetypes:
                newcache[ctype] = self.decodeJPGBuff(jpgbuff, ctype)
                
            self.buff = self.buff[req_len:]
        
        with self.framecondition:
            self.framecondition.notify_all()
            
    
    def decodeJPGBuff(self, jpgbuff, ctype=CONST.CV2):
        if ctype == CONST.CV2:
            return cv2.imdecode(np.frombuffer(jpgbuff, dtype = np.uint8), cv2.COLOR_RGB2BGR)
            
        elif ctype == CONST.IMG:
            return Image.frombytes('RGB',(self.header['width'],self.header['height']), jpgbuff,'jpeg', 'RGB', 'raw')
        elif ctype == CONST.JPG:
            return jpgbuff
    
    
    def cap(self, captype = CONST.CV2, sync=False):
        if sync or self.capcount<=0:
            with self.framecondition:
                self.framecondition.wait(timeout=1.)
        with self.cache.withlatest as cache:
            self.capcount += 1
            capcache = cache.get(captype, None)
            if capcache is not None:
                return capcache
            jpgcache = cache.get(CONST.JPG, None)
            if jpgcache is not None:
                return self.decodeJPGBuff(jpgcache, captype)
        return None
        
    
    def run(self):
        self.stopped = False
        # todo 重复打开端口错误
        self.conn = socket.socket()
        self.conn.connect(('127.0.0.1',self.port)) 
        # todo 判断端口是否正确连接并采取相应措施
        self.__parseHeader()
        
        while not(self.stopped):
            self.__parseFrame()
            
        self.conn.close()
        
        
    def stop(self):
        self.stopped = True
        self.conn.shutdown(socket.SHUT_RDWR)
        '''
        socket.shutdown(how)
        关闭一半或全部的连接。如果 how 为 SHUT_RD，则后续不再允许接收。如果 how 为 SHUT_WR，则后续不再允许发送。如果 how 为 SHUT_RDWR，则后续的发送和接收都不允许。
        '''
        self.join()
        
        
    
class _Minicap:
    
    
    def __init__(self, device_id=None, autorotation=True, **params):
        self.device = device.getDevice(device_id)
        self.id = self.device.id
        self.params = dict(CONST.DEFAULTPARAMS, **params)
        self.isinstalled = False
        self.isensured = False
        self.server = None
        self.autorotation = autorotation
        self.orientation = None
        self.screensize = None
        self.client = None
        self.stopping = False
        if autorotation:
            self.device.supervisorAdd('Orientation')
        
        
    def install(self, fix_mumu=True):
        if self.isinstalled:
            return self.params['remote_bin_path'], self.params['remote_dir']
        print('正在部署minicap。。。', end='')
        dev = self.device
        self.__stopServer()
        abi = dev.abi
        sdk = dev.sdk
        if fix_mumu and sdk == 32 and abi == "x86_64":
            abi = "x86"
        binName = 'minicap' if int(sdk)>=16 else 'minicap-nopie'
        libName = 'minicap.so'
        binPath = os.path.join(CONST.PREBUILDDIR, f'{abi}/bin/{binName}')
        libPath = os.path.join(CONST.PREBUILDDIR, f'{abi}/lib/android-{sdk}/{libName}')
        remoteDir = dev.tmpdir
        remoteBinPath = dev.pushTmp(binPath)
        dev.pushTmp(libPath)
        dev.shell(f'chmod 777 {remoteBinPath}')
        self.params['remote_bin_path'] = remoteBinPath
        self.params['remote_dir'] = remoteDir
        print('ok')
        self.isinstalled = True
        return remoteBinPath, remoteDir
    
    
    def updateParams(self, **params):
        thesame = [v==self.params.get(k,None) for k,v in params.items()]
        if len(thesame)>0 and not all(thesame):
            self.params = dict(self.params, **params)
            self.restart()
    
    
    @property
    def isServerRun(self):
        return self.server and self.server.poll() is None
        
    
    def start(self):
        self.__startServer()
        self.__startClient()
        
        
    def stop(self):
        # 同时只能执行一个关闭过程
        if self.stopping:
            if self.server is not None:
                self.server.wait()
            self.stopping = False
            return
            
        self.stopping = True
        self.__stopClient()
        self.__stopServer()
        self.stopping = False
        
    def close(self):
        self.stop()
        
        
    def restart(self):
        self.stop()
        self.start()
        
    
    def cap(self, captype=CONST.CV2, sync = False):
        if self.stopping:
            self.server.wait()
        if self.autorotation and self.device.supervisor.params['Orientation']!=self.orientation:
            self.restart()
        if self.client is None or not(self.client.is_alive()):
            self.start()
        return self.client.cap(captype, sync)
    
    
    def __startClient(self):
        if self.client is not None and self.client.is_alive():
            return 
        self.client = _Client(self.id, self.params['port'], self.params['cache_size'], cache_types = self.params['cache_types'])
        self.client.start()
        
        
    def __stopClient(self):
        if self.client is None or not(self.client.is_alive()):
            return
        #self.client.stopped = True
        self.client.stop()
        #self.client.join()
    

    def __stopServer(self):
        self.device.shell('pkill minicap')
        if self.server is None:
            return
        self.server.terminate()
        self.server.wait()
    
    
    def __startServer(self):
        if self.isServerRun:
            return 
        remoteBinPath, remoteDir = self.install()
        self.orientation = self.device.getOrientation()
        rotation = self.orientation*90
        self.screensize = self.device.getWMSize()
        screenSize = '%dx%d'%(self.screensize['width'], self.screensize['height'])
        maxfps = self.params['maxfps']
        jpegQuality = self.params['jpeg_quality']
        port = self.params['port']
        cmd = f'LD_LIBRARY_PATH={remoteDir} {remoteBinPath} -P {screenSize}@{screenSize}/{rotation} -r {maxfps} -Q {jpegQuality}'

        self.server = self.device.shell(cmd, readout=False)
        
        while True:
            if 'jpg encoder' in self.server.stderr.readline().decode().lower():
                break
                
        self.device.adb(f'forward tcp:{port} localabstract:minicap')
        
      


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages, Command
from sys import platform as _platform
from shutil import rmtree
import sys
import os

here = os.path.abspath(os.path.dirname(__file__))
NAME = 'anhelper'
REQUIRES_PYTHON = '>=3.0.0'
REQUIRED_DEP = ['adbutils', 'opencv-python', 'flask', 'socketman']
readme = ''
about = {}
history = 'None'
with open(os.path.join(here, 'anhelper', 'version.py')) as f:
    exec(f.read(), about)

with open(os.path.join(here, "README.md"), "rb") as readme_file:
    readme = readme_file.read().decode("UTF-8")

#with open("HISTORY.rst", "rb") as history_file:
#    history = history_file.read().decode("UTF-8")

# OS specific settings
SET_REQUIRES = []
if _platform == "linux" or _platform == "linux2":
   # linux
   print('linux')
elif _platform == "darwin":
   # MAC OS X
   SET_REQUIRES.append('py2app')

required_packages = find_packages()
#required_packages.append('')

APP = [NAME + '.py']
OPTIONS = {
    'argv_emulation': True #,
    #'iconfile': 'icons/app.icns'
}

setup(
    app=APP,
    name=NAME,
    version=about['__version__'],
    description="screencap, touch, remote control for android devices via adb",
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/markdown',
    author="zlols",
    author_email='zlols@foxmail.com',
    url='https://github.com/zlolss/AnHelper',
    python_requires=REQUIRES_PYTHON,
    #package_dir={'': '.'},
    packages=required_packages,
    entry_points={
        'console_scripts': [
            #'testrun=testrun:__main__'
            'anhelper-demo = anhelper.demo:main'
        ]
    },
    include_package_data=True,
    install_requires=REQUIRED_DEP,
    license="MIT license",
    zip_safe=False,
    keywords='test',
    classifiers=[
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: MIT License",
        #'Natural Language :: English',
        #'Programming Language :: Python :: 3',
        #'Programming Language :: Python :: 3.3',
        #'Programming Language :: Python :: 3.4',
        #'Programming Language :: Python :: 3.5',
        #'Programming Language :: Python :: 3.6',
        #'Programming Language :: Python :: 3.7',
    ],
    #package_data={'data/predefined_classes.txt': ['data/predefined_classes.txt']},
    options={'py2app': OPTIONS},
    setup_requires=SET_REQUIRES#,
    # $ setup.py publish support.
    #
    #cmdclass={
    #    'upload': UploadCommand,
    #}
)



# todo: ---------
import threading, time

class ParamSupervisor(threading.Thread):
    '''
    调用father对应getParam方法获取参数
    '''
    def __init__(self, father, interval=1., stop_callback=lambda :None):
        super().__init__()
        self.father = father
        self.params = {}
        self.interval = interval
        self.stopped = False
        self.stopcallback = stop_callback
        self.setDaemon(True)
        
        
    def getMethodName(self, param):
        return f'get{param}'
        
    def addParam(self, param):
        if param in self.params:
            return
        pname = self.getMethodName(param)
        if pname not in self.father.__dir__():
            print(f'get param:{pname} failed')
            return 
        self.params[param] = None
        
    def updateParam(self, param):
        try:
            pname = self.getMethodName(param)
            self.params[param] = eval(f'self.father.{pname}()')
        except:
            print('参数获取错误，supervisor正在停止')
            self.stopped = True
            self.stopcallback()
        
        
    def run(self):
        while not(self.stopped):
            params = list(self.params.keys())
            for param in params:
                self.updateParam(param)
            time.sleep(self.interval)


class Cache:

    def __init__(self, maxsize = 10):
        self.size = maxsize
        self.seek = 0
        self.latestseek = 0
        self.newseek = 0
        self.data = [{} for i in range(maxsize+1)]

    
    def nextSeek(self, seek):
        return seek+1 if seek<self.size else 0
    
    
    class GetCache:
        def __init__(self, father, seek, isnew = False):
            self.father = father
            self.seek = seek
            self.isnew = isnew
            
        def __enter__(self):
            #self.father.locked.add(self.seek)
            return self.father.data[self.seek]
            
        def __exit__(self, exc_type, exc_value, traceback):
            #self.father.locked.remove(self.seek)
            if self.isnew:
                self.father.latestseek = self.seek
            pass
            
    
    @property
    def withnew(self):
        # 写入
        seek = self.newseek
        self.newseek = self.nextSeek(self.newseek)
        return self.GetCache(self, seek, isnew = True)
        
    
    @property
    def withlatest(self):
        # 读取
        return self.GetCache(self, self.latestseek)
        